__version__ = "0.3.1"

from .experiment import (
    Launcher,
    Reports,
)
from .utilities.loaders import (
    ConfigLoader,
    load_problem,
)

__all__ = [
    "ConfigLoader",
    "Launcher",
    "Reports",
    "load_problem",
]



from collections import defaultdict
from enum import Enum, auto
from typing import Any, Optional

from pydantic import validate_call

from ..utilities.loggers import logger
from ..utilities.tools import tabulate_formats, titled_tabulate

__all__ = ['Actions', 'ClientPackage', 'SyncDataManager']


class Actions(Enum):
    REGISTER = auto()
    QUIT = auto()


class ClientPackage:
    def __init__(self, cid: Optional[int], action: Any):
        self.cid = cid
        self.action = action


class SyncDataManager:
    def __init__(self):
        self._source: dict[int, dict[int, Any]] = defaultdict(dict)
        self._result: dict[int, dict[int, Any]] = defaultdict(dict)

    @validate_call
    def update_src(self, cid: int, version: int, data: Any) -> None:
        """
        Update aggregation source of a client
        """
        self._source[cid][version] = data

    @validate_call
    def update_res(self, cid: int, version: int, data: Any) -> None:
        """
        Update aggregation results of a client
        """
        self._result[cid][version] = data

    @validate_call
    def lts_src_ver(self, cid: int) -> int:
        """
        Return the latest source version of a client
        """
        data = self._source.get(cid, {-1: None})
        return max(data.keys())

    @validate_call
    def lts_res_ver(self, cid: int) -> int:
        """
        Return the latest aggregation result version of a client
        """
        data = self._result.get(cid, {-1: None})
        return max(data.keys())

    @validate_call
    def get_src(self, cid: int, version: int) -> Any:
        """
        - Return the aggregation source
        - Return None if not found
        """
        if cid not in self._source:
            logger.debug(f"cid={cid} not found in source")
            return None
        data = self._source[cid].get(version)
        if data is None:
            logger.debug(f"cid={cid} version={version} not found in source")
        return data

    @validate_call
    def get_res(self, cid: int, version: int) -> Any:
        """
        - Return the aggregation result data
        - Return None if not found
        """
        if cid not in self._result:
            logger.debug(f"cid={cid} not found in result")
            return None
        data = self._result[cid].get(version)
        if data is None:
            logger.debug(f"cid={cid} version={version} not found in result")
        return data

    @property
    def available_src_ver(self) -> int:
        """
        - The latest consistent version uploaded by all clients
        - -1 if no source data is available
        """
        try:
            return min([max(data.keys()) for data in self._source.values()])
        except ValueError:
            return -1

    @property
    def num_clients(self) -> int:
        """
        The number of clients that have uploaded least 1 source data
        """
        return len(self._source)

    @property
    def data_info(self):
        src = {f"C {cid}": [self.lts_src_ver(cid)] for cid in sorted(self._source.keys())}
        res = {f"C {cid}": [self.lts_res_ver(cid)] for cid in sorted(self._result.keys())}
        src_str = titled_tabulate("Aggregation Source", '=',
                                  src, headers='keys', tablefmt=tabulate_formats.rounded_grid) if src else ""
        res_str = titled_tabulate("Aggregation Results", '=',
                                  res, headers='keys', tablefmt=tabulate_formats.rounded_grid) if res else ""
        return (
            f"AvaSrcVer: {self.available_src_ver}\n"
            f"Total Clt: {self.num_clients}"
            f"{src_str}{res_str}"
        )



import subprocess
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager

from rich import box
from rich import progress as rpg
from rich.console import Console, Group
from rich.live import Live
from rich.measure import measure_renderables
from rich.table import Table
from setproctitle import setproctitle
from tabulate import tabulate

from ..framework import Client
from ..utilities.loaders import ExperimentConfig, LauncherConfig
from ..utilities.loggers import logger
from ..utilities.tools import (
    clear_console,
    redirect_warnings,
    terminate_popen,
    titled_tabulate,
)
from ..utilities.tools import tabulate_formats as tf
from .utils import RunSolutions

__all__ = ['Launcher']


class Launcher:
    exp_idx: int
    exp: ExperimentConfig
    clients: list[Client]
    table: Table
    progress: rpg.Progress

    def __init__(self, conf: LauncherConfig):
        self.conf = conf

        # Runtime data
        self._repeat_id = 0
        self._results: list[Client] = []

    def run(self):
        self._setup()
        for self.exp_idx, self.exp in enumerate(self.conf.experiments):  # noqa: B020
            logger.info(f"\n{self.exp}")
            if self.conf.save:
                self._init_root()
                self._create_snapshot()
            with redirect_warnings():
                self._repeating()
        self._remove_bars()
        self.conf.show_summary()

    def _setup(self):
        self.progress = rpg.Progress(
            rpg.TextColumn("[progress.description]{task.description}"),
            rpg.BarColumn(bar_width=None),
            rpg.TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            rpg.MofNCompleteColumn(),
            rpg.TimeRemainingColumn(),
            expand=True,
        )
        setproctitle("AlgClients")

    def _repeating(self):
        self._repeat_id = 0
        self._update_repeat_id()
        while not self._finished:
            self._init_clients()
            self._launch_exp()
            self._save_results()
            self._update_repeat_id()
            time.sleep(1)
        self.exp.success = True

    def _init_clients(self):
        problem = self.exp.problem.initialize()
        clt_params = self.exp.algorithm.params.get('client', {})
        self.clients = [self.exp.algorithm.client(p, **clt_params) for p in problem]

    def _launch_exp(self):
        clear_console()
        with Live(self._panel, console=Console(width=self.width), refresh_per_second=10):
            with self._running_server():
                self._start_clients()

    @property
    def _panel(self) -> Group:
        self._remove_bars()
        self._update_table()
        return Group(self.table, self.progress)

    @property
    def width(self):
        csl = Console(width=1000)
        min_width, _ = measure_renderables(csl, csl.options, [self.table])
        return min_width

    def _remove_bars(self):
        for bar in self.progress.task_ids:
            self.progress.remove_task(bar)

    def _update_table(self):
        curr_rep = self.conf.repeat * self.exp_idx + self._repeat_id
        total_rep = self.conf.total_repeat
        tab_dict = {
            'running': f"{self.exp_idx+1}/{self.conf.n_exp}",
            'repeating': f"{self._repeat_id}/{self.conf.repeat}",
            'progress': f"[{curr_rep}/{total_rep}][{100 * curr_rep / total_rep:.2f}%]",
            'algorithm': self.exp.algorithm.name,
            'problem': self.exp.problem.name,
            'NPD': self.clients[0].problem.npd,
            'clients': len(self.clients),
            'save': self.conf.save
        }

        def mapper(v):
            if v is True:
                return '[green]yes[/green]'
            elif v is False:
                return '[red]no[/red]'
            elif isinstance(v, int):
                return f'[magenta]{v}[/magenta]'
            else:
                return str(v)

        keys = list(tab_dict.keys())
        values = list(tab_dict.values())
        colored_values = list(map(mapper, values))

        tab_csl = Table(box=box.ROUNDED)
        [tab_csl.add_column(k, justify='center') for k in keys]
        tab_csl.add_row(*colored_values)
        self.table = tab_csl

        alignment = ['center'] * len(keys)
        original_tab = {k: [v] for k, v in zip(keys, values)}
        tab_log = tabulate(original_tab, headers='keys', tablefmt='rounded_grid', colalign=alignment)
        logger.info(f"\n{tab_log}")

    @contextmanager
    def _running_server(self):
        server = self.exp.algorithm.server
        kwargs = self.exp.algorithm.params.get('server', {})

        cmd = [
            sys.executable, "-c",
            "from pyfmto.utilities.loggers import logger; "
            "from pyfmto.utilities.loaders import add_sources; "
            "from importlib import import_module; "
            f"add_sources({self.conf.sources}); "
            f"module = import_module('{server.__module__}'); "
            f"logger.setLevel('{self.conf.loglevel}'); "
            f"srv = module.{server.__name__}(**{kwargs!r}); "
            f"srv.start()"
        ]

        process = subprocess.Popen(
            cmd,
            stdin=subprocess.DEVNULL,
            stdout=None,
            stderr=None
        )
        logger.info("Server started.")
        time.sleep(3)
        try:
            yield process
        finally:
            terminate_popen(process)
            logger.debug("Server terminated.")

    def _start_clients(self):
        n_clients = len(self.clients)
        for c in self.clients:
            bar = self.progress.add_task(c.name, total=c.fe_max)
            c.progress = self.progress
            c.bar = bar
        pool = ThreadPoolExecutor(max_workers=n_clients)
        futures = [pool.submit(c.start) for c in self.clients]
        pool.shutdown(wait=True)
        self._results = [fut.result() for fut in futures]

    def _init_root(self):
        self.exp.init_root()

    def _create_snapshot(self):
        if self.conf.snapshot:
            self.exp.create_snapshot(self.conf.packages)

    def _save_verbose(self):
        tables = {clt.id: clt for clt in self._results}
        info_str = ''
        for cid in sorted(tables.keys()):
            clt = tables[cid]
            tab = titled_tabulate(
                f"{clt.name} {clt.problem.name}({clt.dim}D)",
                '=', clt.rounds_info, headers='keys', tablefmt=tf.rounded_grid
            )
            info_str = f"{info_str}{tab}\n"
        res_file = self.exp.result_name(self._repeat_id)
        log_dir = res_file.with_name('verbose')
        log_name = res_file.with_suffix('.log').name
        log_dir.mkdir(parents=True, exist_ok=True)
        with open(log_dir / log_name, 'w') as f:
            f.write(info_str)

    def _save_results(self):
        if self.conf.save:
            run_solutions = RunSolutions()
            for clt in self._results:
                run_solutions.update(clt.id, clt.solutions)
            run_solutions.to_msgpack(self.exp.result_name(self._repeat_id))
            if self.conf.verbose:
                self._save_verbose()

    def _update_repeat_id(self):
        if self.conf.save:
            self._repeat_id = self.exp.num_results + 1
        else:
            self._repeat_id += 1

    @property
    def _finished(self):
        return self._repeat_id > self.conf.repeat



import copy
import importlib
import inspect
import os
import shutil
import textwrap
from collections import defaultdict
from datetime import datetime
from itertools import chain, product
from pathlib import Path
from textwrap import indent
from typing import Any, Literal, Union

import psutil
import tabulate
from deepdiff import DeepDiff
from pydantic import BaseModel, ConfigDict
from rich import box
from rich.console import Console
from rich.table import Table
from ruamel.yaml import CommentedMap

from ..framework.client import Client
from ..framework.server import Server
from ..problem import MultiTaskProblem
from .io import dumps_yaml, load_yaml, parse_yaml
from .loggers import logger
from .tools import add_sources, clear_console, get_os_name, tabulate_formats, titled_tabulate

__all__ = [
    'AlgorithmData',
    'ConfigLoader',
    'ExperimentConfig',
    'LauncherConfig',
    'ProblemData',
    'ReporterConfig',
    'load_problem',
]


def recursive_to_pure_dict(data: Any) -> dict[str, Any]:
    """
    Recursively convert nested dict and CommentedMap objects to a pure Python
    dictionary to avoid YAML serialization issues.
    """
    if isinstance(data, (dict, CommentedMap)):
        for k, v in data.items():
            data[k] = recursive_to_pure_dict(v)
    else:
        return data
    return data


def combine_params(params: dict[str, Union[Any, list[Any]]]) -> list[dict[str, Any]]:
    values = []
    for v in params.values():
        if isinstance(v, list):
            values.append(v)
        else:
            values.append([v])
    result = []
    for combination in product(*values):
        result.append(dict(zip(params.keys(), combination)))
    return result


def load_problem(name: str, config: Union[str, Path] = 'config.yaml', **kwargs) -> MultiTaskProblem:
    conf = ConfigLoader(config)
    prob = conf.problems.get(name, ProblemData(name, []))
    if not prob.available:
        raise ValueError(f"Problem '{name}' is not available")
    prob.params_update = kwargs
    return prob.initialize()


class AlgorithmData:
    client: type[Client]
    server: type[Server]

    def __init__(self, name: str, paths: list[str]):
        self.name_orig = name
        self.name_alias = ''
        self.paths = paths
        self.params_default: dict[str, dict[str, Any]] = {}
        self.params_update: dict[str, dict[str, Any]] = {}
        self.module_detail: dict[str, list] = defaultdict(list)
        self.__load()
        if self.available:
            self.__parse_default_params()

    @property
    def available(self):
        return True in self.module_detail['pass']

    def verbose(self):
        return {k: self.module_detail[k] for k in ['name', 'pass', 'path', 'msg']}

    def __load(self):
        check_res: dict[str, list] = defaultdict(list)

        for path in self.paths:
            try:
                module = importlib.import_module(path)
                clt, srv = self._extract_client_server(module)
                msg = self._build_check_message(clt, srv)
                check_pass = clt is not None and srv is not None

                # Only set client/server once (first valid module)
                if check_pass and not hasattr(self, 'client'):
                    self.client = clt
                    self.server = srv

            except Exception as e:
                clt, srv = None, None
                msg = f"Exception: {e!s}"
                check_pass = False

            check_res['name'].append(self.name)
            check_res['pass'].append(check_pass)
            check_res['path'].append(path)
            check_res['client'].append(clt)
            check_res['server'].append(srv)
            check_res['msg'].append(msg)

        self.module_detail = check_res

    def _extract_client_server(self, module):
        clt, srv = None, None
        for attr_name in dir(module):
            if attr_name.startswith('__'):
                continue
            attr = getattr(module, attr_name)
            is_client = inspect.isclass(attr) and issubclass(attr, Client) and attr is not Client
            is_server = inspect.isclass(attr) and issubclass(attr, Server) and attr is not Server
            if is_client:
                clt = attr
            if is_server:
                srv = attr
            if clt and srv:
                break  # Early exit: both found
        return clt, srv

    def _build_check_message(self, clt, srv):
        msg_parts = []
        if not clt:
            msg_parts.append("The subclass of 'Client' not found.")
        if not srv:
            msg_parts.append("The subclass of 'Server' not found.")
        return '\n'.join(msg_parts)

    def __parse_default_params(self):
        c_doc = self.client.__doc__ if self.client else None
        s_doc = self.server.__doc__ if self.server else None
        c_args = parse_yaml(c_doc) if c_doc else {}
        s_args = parse_yaml(s_doc) if s_doc else {}
        if c_args:
            self.params_default.update({'client': c_args})
        if s_args:
            self.params_default.update({'server': s_args})

    def copy(self) -> 'AlgorithmData':
        return copy.deepcopy(self)

    @property
    def params(self) -> dict[str, dict[str, Any]]:
        kwargs = copy.deepcopy(self.params_default)
        for k in ['client', 'server']:
            for k2, v2 in self.params_update.get(k, {}).items():
                if k not in kwargs:
                    kwargs[k] = {}
                kwargs[k][k2] = v2
        return kwargs

    @property
    def params_diff(self) -> str:
        return DeepDiff(self.params_default, self.params).pretty()

    @property
    def name(self) -> str:
        return self.name_orig if not self.name_alias else self.name_alias

    @property
    def params_yaml(self) -> str:
        if self.params_default:
            return dumps_yaml(self.params_default)
        else:
            return f"Algorithm '{self.name}' no configurable parameters."


class ProblemData:
    problem: type[MultiTaskProblem]

    def __init__(self, name, paths: list[str]):
        self.name_orig = name
        self.paths = paths
        self.params_default: dict[str, Any] = {
            'npd': 1,
            'random_ctrl': 'weak',
            'seed': 123,
        }
        self.params_update: dict[str, Any] = {}
        self.module_detail: dict[str, list] = defaultdict(list)
        self.__load()
        if self.available:
            self.__parse_default_params()

    @property
    def available(self):
        return hasattr(self, 'problem')

    def verbose(self):
        return {k: self.module_detail[k] for k in ['name', 'pass', 'path', 'msg']}

    def __load(self):
        check_res: dict[str, list] = defaultdict(list)

        for path in self.paths:
            try:
                module = importlib.import_module(path)
                problem = self._find_problem_class(module)
                if problem is not None:
                    if not hasattr(self, 'problem'):
                        self.problem = problem
                    check_pass = True
                    msg = ""
                else:
                    check_pass = False
                    msg = "The subclass of 'MultiTaskProblem' not found."
            except Exception as e:
                problem = None
                check_pass = False
                msg = str(e)

            check_res['name'].append(self.name_orig)
            check_res['pass'].append(check_pass)
            check_res['path'].append(path)
            check_res['problem'].append(problem)
            check_res['msg'].append(msg)

        self.module_detail = check_res

    def _find_problem_class(self, module):
        for attr_name in dir(module):
            if attr_name.startswith('__'):
                continue
            attr = getattr(module, attr_name)
            if (
                    inspect.isclass(attr)
                    and issubclass(attr, MultiTaskProblem)
                    and attr is not MultiTaskProblem
            ):
                return attr
        return None

    def __parse_default_params(self):
        p_doc = self.problem.__doc__
        self.params_default.update(parse_yaml(p_doc))

    def copy(self) -> 'ProblemData':
        return copy.deepcopy(self)

    @property
    def params(self) -> dict[str, Any]:
        params = copy.deepcopy(self.params_default)
        params.update(self.params_update)
        dim = params.get('dim', 0)
        if dim > 0:
            if 'fe_init' not in params:
                params.update(fe_init=5 * dim)
            if 'fe_max' not in params:
                params.update(fe_max=11 * dim)
        return params

    @property
    def params_diff(self) -> str:
        return DeepDiff(self.params_default, self.params).pretty()

    def initialize(self) -> MultiTaskProblem:
        if self.available:
            return self.problem(**self.params)
        else:
            raise ValueError(f"Problem {self.name_orig} not available.")

    @property
    def name(self) -> str:
        name_with_dim = f"{self.name_orig}_{self.task_str}_{self.dim_str}"
        name_no_dim = f"{self.name_orig}_{self.task_str}"
        return name_with_dim if self.dim > 0 else name_no_dim

    @property
    def npd(self) -> int:
        return self.params.get('npd', 0)

    @property
    def npd_str(self) -> str:
        return f"NPD{self.npd}" if self.npd > 0 else ""

    @property
    def dim(self) -> int:
        return self.params.get('dim', 0)

    @property
    def dim_str(self) -> str:
        if self.dim > 0:
            return f"{self.dim}D"
        else:
            return ""

    @property
    def n_task(self) -> int:
        if self.available:
            return len(self.problem(**{'_init_solutions': False}, **self.params))
        else:
            raise ValueError(f"Problem {self.name_orig} not available.")

    @property
    def task_str(self) -> str:
        return f"{self.n_task}T"

    @property
    def params_yaml(self) -> str:
        if self.params_default:
            return dumps_yaml(self.params_default)
        else:
            return f"Problem '{self.name}' no configurable parameters."


class ExperimentConfig:

    def __init__(
            self,
            algorithm: AlgorithmData,
            problem: ProblemData,
            root: str
    ):
        self.algorithm = algorithm
        self.problem = problem
        self._root = Path(root)
        self.success = False

    @property
    def root(self) -> Path:
        return self._root / self.algorithm.name / self.problem.name / self.problem.npd_str

    @property
    def code_dest(self) -> Path:
        today = datetime.now().strftime("%Y-%m-%d")
        return self.root.parent / "snapshot" / f"code {today}"

    @property
    def markdown_dest(self) -> Path:
        today = datetime.now().strftime("%Y-%m-%d")
        return self.root.parent / "snapshot" / f"markdown {today}.md"

    @property
    def prefix(self) -> str:
        fe_init = self.problem.params.get('fe_init')
        fe_max = self.problem.params.get('fe_max')
        seed = self.problem.params.get('seed')
        fe_i = "" if fe_init is None else f"FEi{fe_init}_"
        fe_m = "" if fe_max is None else f"FEm{fe_max}_"
        seed = "" if seed is None else f"Seed{seed}_"
        return f"{fe_i}{fe_m}{seed}"

    def result_name(self, file_id: int):
        return self.root / f"{self.prefix}Rep{file_id:02d}.msgpack"

    @property
    def num_results(self) -> int:
        if not self.root.exists():
            return 0
        prefix = self.result_name(0).name.split('Rep')[0]
        suffix = '.msgpack'
        results = [f for f in os.listdir(self.root) if f.startswith(prefix) and f.endswith(suffix)]
        return len(results)

    @property
    def params_dict(self) -> dict[str, Any]:
        data = {
            'algorithm': {
                self.algorithm.name: {
                    'base': self.algorithm.name_orig,
                    'params': self.algorithm.params,
                    'default': self.algorithm.params_default,
                    'update': self.algorithm.params_update,
                },
            },
            'problem': {
                self.problem.name: {
                    'params': self.problem.params,
                    'default': self.problem.params_default,
                    'update': self.problem.params_update,
                }
            }
        }
        return recursive_to_pure_dict(data)

    @staticmethod
    def desc_sys():
        from pyfmto.utilities.tools import get_cpu_model
        data = {
            'OS': get_os_name(),
            "CPU": get_cpu_model(),
            "MEM": f"{round(psutil.virtual_memory().total / (1024 ** 3), 1)} GB"
        }
        return '\n'.join([f"- {k}: {v}" for k, v in data.items()])

    @staticmethod
    def desc_env(packages: list[str]):
        import platform

        from pyfmto.utilities.tools import get_pkgs_version
        data = {
            'python': platform.python_version(),
            **get_pkgs_version(packages)
        }
        return '\n'.join([f"- {k}: `{v}`" for k, v in data.items()])

    def create_snapshot(self, packages: list[str]):
        date = datetime.now().strftime("%Y-%m-%d")
        time = datetime.now().strftime("%H:%M:%S")
        template = f"""
            # Experiment information

            The experiment was performed on `{date}` at `{time}`

            ---

            ## system

            <system>

            ## environment

            <environment>

            ## Configuration

            ### Algorithm

            ``` yaml
            <algorithm>
            ```

            ### Problem

            ``` yaml
            <problem>
            ```
        """
        md_str = textwrap.dedent(template)
        md_str = md_str.replace('<system>', self.desc_sys())
        md_str = md_str.replace('<environment>', self.desc_env(packages))
        md_str = md_str.replace('<algorithm>', dumps_yaml(self.params_dict['algorithm']))
        md_str = md_str.replace('<problem>', dumps_yaml(self.params_dict['problem']))

        file_src = inspect.getfile(self.algorithm.client)
        for p in Path(file_src).resolve().parents:
            if p.parent.name == 'algorithms' and not self.code_dest.exists():
                shutil.copytree(p, self.code_dest)
        if not self.markdown_dest.exists():
            with open(self.markdown_dest, 'w') as f:
                f.write(md_str)

    def init_root(self):
        self.root.mkdir(parents=True, exist_ok=True)

    def __str__(self):
        tab = {
            'Algorithm': [self.algorithm.name],
            'Problem': [self.problem.name],
            'NPD': [self.problem.npd_str],
            'Dimension': [self.problem.dim_str],
        }
        return titled_tabulate("Experiment", '=', tab, tablefmt='rounded_grid')

    def __repr__(self):
        info = [
            f"Alg({self.algorithm.name})",
            f"Prob({self.problem.name})",
            f"NPD({self.problem.params['npd']})",
            f"Dim({self.problem.params.get('dim', '-')})",
        ]

        return ' '.join(info)


class LauncherConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    sources: list[str]
    results: str
    repeat: int
    seed: int
    save: bool
    loglevel: str
    snapshot: bool
    verbose: bool
    packages: list[str]
    algorithms: list[str]
    problems: list[str]
    experiments: list[ExperimentConfig] = []

    def show_summary(self):
        tab = Table(
            title="Experiments Summary",
            title_justify="center",
            box=box.ROUNDED,
        )
        tab.add_column('Algorithm', justify='center', style="cyan")
        tab.add_column('Original', justify='center', style="cyan")
        tab.add_column('Problem', justify='center', style="magenta")
        tab.add_column('NPD', justify='center', style="yellow")
        tab.add_column('Success', justify='center')

        for exp in self.experiments:
            tab.add_row(
                exp.algorithm.name,
                exp.algorithm.name_orig,
                exp.problem.name,
                exp.problem.npd_str,
                '[green]yes[/green]' if exp.success else '[red]no[/red]'
            )
        clear_console()
        Console().print(tab)

    @property
    def n_exp(self) -> int:
        return len(self.experiments)

    @property
    def total_repeat(self) -> int:
        return self.n_exp * self.repeat


class ReporterConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    results: str
    algorithms: list[list[str]]
    problems: list[str]
    formats: list[str]
    params: dict[str, Any] = {}
    experiments: list[ExperimentConfig] = []
    groups: list[tuple[list[str], str, str]] = []

    @property
    def root(self) -> Path:
        return Path(self.results)


class ConfigLoader:
    """
    launcher:
        sources: []           # load alg/prob in these directory
        results: out/results  # [optional] save results to this directory
        repeat: 2             # [optional] repeat each experiment for this number of times
        seed: 123             # [optional] random seed
        save: true            # [optional] save results to disk
        loglevel: INFO        # [optional] log level [CRITICAL, ERROR, WARNING, INFO, DEBUG], default INFO
        snapshot: true        # [optional] If create snapshot of the experiment
        verbose: false        # [optional] Save detailed information for each repeat run
        packages: []          # [optional] Record the version of these packages
        algorithms: []        # run these algorithms
        problems: []          # run each algorithm on these problems
    reporter:
        formats: [excel]      # [optional] generate these reports
    """

    def __init__(self, config: Union[str, Path] = 'config.yaml'):
        self.config_default = parse_yaml(self.__class__.__doc__)
        self.config_update = load_yaml(config, ignore_errors=True)
        self.config = copy.deepcopy(self.config_default)
        self.merge_global_config_from_updates()
        self.fill_reporter_config_from_launcher()
        self.algorithms: dict[str, AlgorithmData] = {}
        self.problems: dict[str, ProblemData] = {}
        add_sources(self.sources)
        self.__list_sources()

    @property
    def sources(self) -> list[str]:
        return self.config['launcher']['sources']

    def merge_global_config_from_updates(self):
        for key, value in self.config_update.items():
            if key in self.config:
                self.config[key].update(value)
            else:
                self.config[key] = value
        cwd = str(Path().cwd().resolve())
        if cwd not in self.config['launcher']['sources']:
            self.config['launcher']['sources'].append(cwd)
        logger.setLevel(self.config['launcher']['loglevel'])

    def fill_reporter_config_from_launcher(self) -> None:
        launcher_params = self.config['launcher']
        for key in ['results', 'problems', 'algorithms']:
            if key in self.config['reporter']:
                continue
            if key == 'algorithms':
                self.config['reporter'][key] = [launcher_params[key]]
            else:
                self.config['reporter'][key] = launcher_params[key]

    def __list_sources(self):
        for target in ['algorithms', 'problems']:
            source_indices: dict[str, list[str]] = defaultdict(list)
            for path in self.sources:
                target_dir = Path(path).resolve() / target
                if target_dir.exists():
                    for name in os.listdir(target_dir):
                        sub_dir = target_dir / name
                        if sub_dir.is_dir() and not name.startswith(('.', '_')):
                            source_indices[name].append('.'.join(sub_dir.parts[-3:]))
            if target == 'algorithms':
                for name, paths in source_indices.items():
                    self.algorithms[name] = AlgorithmData(name, paths)
            else:
                for name, paths in source_indices.items():
                    self.problems[name] = ProblemData(name, paths)
        logger.debug(self.show_sources('algorithms'))
        logger.debug(self.show_sources('problems'))

    def show_sources(self, target: Literal['algorithms', 'problems'], print_it: bool = False) -> str:
        res: dict[str, list[Any]] = defaultdict(list)
        src_str = '\n'.join(self.sources)
        target_values = list(getattr(self, target).values())
        if len(target_values) == 0:
            summary = f"No {target} found in {src_str}"
        else:
            dicts = [data.verbose() for data in getattr(self, target).values()]
            keys = dicts[0].keys()

            for k in keys:
                for d in dicts:
                    res[k] += d[k]

            summary = f"Found {sum(res['pass'])} available (total {len(res['pass'])}) {target} in\n{src_str}"
        if print_it:
            print(summary)
            from pyfmto.utilities.tools import print_dict_as_table
            print_dict_as_table(res)
        tab = tabulate.tabulate(res, headers='keys', tablefmt=tabulate_formats.rounded_grid) if res != {} else ''
        return f"{summary}\n{tab}"

    @property
    def launcher(self) -> LauncherConfig:
        self.check_config_issues('launcher')
        conf = LauncherConfig(**self.config['launcher'])
        logger.setLevel(conf.loglevel)
        algorithms = self.gen_alg_list(conf.algorithms)
        problems = self.gen_prob_list(conf.problems)
        logger.debug(f"algorithms: {[alg.name for alg in algorithms]}")
        logger.debug(f"problems: {[prob.name for prob in problems]}")
        conf.experiments = [ExperimentConfig(alg, prob, conf.results) for alg, prob in product(algorithms, problems)]
        return conf

    @property
    def reporter(self) -> ReporterConfig:
        self.check_config_issues('reporter')
        conf = ReporterConfig(**self.config['reporter'])
        alg_names = list(set(chain.from_iterable(conf.algorithms)))
        algorithms = [AlgorithmData(name, []) for name in alg_names]
        problems = self.gen_prob_list(conf.problems)
        conf.experiments = [
            ExperimentConfig(alg, prob, conf.results)
            for alg, prob in product(algorithms, problems)
        ]
        conf.groups = [
            (algs, prob.name, prob.npd_str)
            for algs, prob in list(product(conf.algorithms, problems))
        ]
        return conf

    def gen_alg_list(self, names: list[str]) -> list[AlgorithmData]:
        algorithms: list[AlgorithmData] = []
        for name_alias in names:
            alg_params = self.config.get('algorithms', {}).get(name_alias, {})
            alg_name = alg_params.pop('base', name_alias)
            if alg_name not in self.algorithms or not self.algorithms[alg_name].available:
                logger.error(f"Algorithm '{alg_name}' is not available")
                continue
            alg_data = self.algorithms[alg_name].copy()
            alg_data.name_alias = name_alias
            alg_data.params_update = alg_params
            algorithms.append(alg_data)
        return algorithms

    def gen_prob_list(self, names: list[str]) -> list[ProblemData]:
        problems: list[ProblemData] = []
        for n in names:
            if n not in self.problems or not self.problems[n].available:
                logger.error(f"Problem {n} is not available.")
                continue
            prob_params = self.config.get('problems', {}).get(n, {})
            params_variations = combine_params(prob_params)
            for params in params_variations:
                prob_data = self.problems[n].copy()
                prob_data.params_update = params
                problems.append(prob_data)
        return problems

    def check_config_issues(self, name: Literal['launcher', 'reporter']) -> None:
        if name == 'launcher':
            issues = self.check_launcher_config()
        else:
            issues = self.check_reporter_config()
        if issues:
            detail = indent('\n'.join(issues), ' ' * 4)
            msg = f"{name.title()} configuration issues:\n{detail}"
            logger.error(msg)
            raise ValueError(msg)

    def check_launcher_config(self) -> list[str]:
        issues = []
        launcher = self.config['launcher']
        if not launcher.get('results'):
            issues.append("No results directory specified in launcher.")
        if launcher.get('repeat') <= 0:
            issues.append("Invalid repeat number specified in launcher. Must be greater than 0.")
        if not isinstance(launcher.get('save'), bool):
            issues.append("Invalid save option specified in launcher. Must be True or False.")
        if not launcher.get('algorithms'):
            issues.append("No algorithms specified in launcher.")
        if not launcher.get('problems'):
            issues.append("No problems specified in launcher.")
        return issues

    def check_reporter_config(self) -> list[str]:
        issues = []
        reporter = self.config['reporter']
        if not reporter.get('results'):
            issues.append("No results directory specified in reporter or launcher.")
        if not reporter.get('algorithms', []):
            issues.append("No algorithms specified in reporter or launcher.")
        else:
            validate_values: list[list[str]] = []
            for item in reporter['algorithms']:
                if isinstance(item, str) and item:
                    validate_values.append([item])
                elif isinstance(item, list) and item:
                    validate_values.append(item)
                else:
                    issues.append(f"Invalid value [type:{type(item)}, value:{item}] specified in reporter.")
                reporter['algorithms'] = validate_values
        return issues



import argparse
from pathlib import Path
from typing import Literal, cast

from ..experiment import Launcher, Reports, list_report_formats, show_default_conf
from .loaders import AlgorithmData, ConfigLoader, ProblemData
from .tools import add_sources, matched_str_head


def main() -> None:
    add_sources([str(Path.cwd())])

    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        '-c', '--config', type=str, default='config.yaml',
        help='Path to configuration file (default: config.yaml)'
    )

    parser = argparse.ArgumentParser(
        description='PyFMTO: Python Library for Federated Many-task Optimization Research'
    )
    subparsers = parser.add_subparsers(dest='command', help='Available commands')

    # Run/Report command
    subparsers.add_parser('run', parents=[global_parser], help='Run experiments')
    subparsers.add_parser('report', parents=[global_parser], help='Generate reports')

    # List command
    list_p = subparsers.add_parser('list', parents=[global_parser], help='List available options')
    list_p.add_argument(
        'name', type=str, help='Name of the option to list'
    )

    # Show command
    show_p = subparsers.add_parser('show', parents=[global_parser], help='Show default configurations')
    show_p.add_argument(
        'name', type=str,
        help="Name of the configuration to show, any things that can be list by the 'list' command"
    )
    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return

    conf = ConfigLoader(config=args.config)

    handlers = {
        "run": lambda: _handle_run(conf),
        "report": lambda: _handle_report(conf),
        "list": lambda: _handle_list(args, conf),
        "show": lambda: _handle_show(args, conf),
    }
    handlers[args.command]()


def _handle_run(conf: ConfigLoader):
    launcher = Launcher(conf=conf.launcher)
    launcher.run()


def _handle_report(conf: ConfigLoader):
    reports = Reports(conf=conf.reporter)
    reports.generate()


def _handle_list(args, conf: ConfigLoader):
    full_name = matched_str_head(args.name, ['problems', 'algorithms', 'reports'])
    if full_name in ('problems', 'algorithms'):
        conf.show_sources(cast(Literal['algorithms', 'problems'], full_name), print_it=True)
    elif full_name == 'reports':
        list_report_formats(print_it=True)
    else:
        print(f"Unknown list option: {args.name}")


def _handle_show(args, conf: ConfigLoader):
    if '.' not in args.name:
        print("Error: name must be in format 'group.item' (e.g., problems.MyProblem)")
        return

    t, v = args.name.split('.', 1)
    full_name = matched_str_head(t, ['problems', 'algorithms', 'reports'])

    if full_name == 'problems':
        prob = conf.problems.get(v, ProblemData(v, []))
        print(prob.params_yaml)
    elif full_name == 'algorithms':
        alg = conf.algorithms.get(v, AlgorithmData(v, []))
        print(alg.params_yaml)
    elif full_name == 'reports':
        show_default_conf(v)
    else:
        print(f"No matched group for '{t}'. Available: problems, algorithms, reports")



import sys
from pathlib import Path
from unittest.mock import Mock, patch

from pyfmto.experiment import list_report_formats
from pyfmto.utilities.cli import main
from tests.helpers import PyfmtoTestCase, gen_code
from tests.helpers.generators import gen_config


class TestMainFunction(PyfmtoTestCase):

    def setUp(self):
        self.save_sys_env()
        self.tmp_file = 'temp_conf.yaml'

    def tearDown(self):
        self.restore_sys_env()

    @patch('pyfmto.utilities.cli.Launcher')
    @patch('pyfmto.utilities.cli.ConfigLoader')
    def test_run_command(self, mock_loader, mock_launcher):
        mock_launcher_instance = Mock()
        mock_launcher.return_value = mock_launcher_instance
        mock_loader.return_value = Mock()

        test_args = ['pyfmto', 'run', '--config', self.tmp_file]
        with patch.object(sys, 'argv', test_args):
            main()

        mock_loader.assert_called_once_with(config=self.tmp_file)
        mock_launcher.assert_called_once()
        mock_launcher_instance.run.assert_called_once()

        with patch.object(sys, 'argv', ['pyfmto']):
            main()

    @patch('pyfmto.utilities.cli.Reports')
    @patch('pyfmto.utilities.cli.ConfigLoader')
    def test_report_command(self, mock_loader, mock_reports):
        mock_reports_instance = Mock()
        mock_reports.return_value = mock_reports_instance
        mock_loader.return_value = Mock()

        test_args = ['pyfmto', 'report', '--config', self.tmp_file]
        with patch.object(sys, 'argv', test_args):
            main()

        mock_loader.assert_called_once_with(config=self.tmp_file)
        mock_reports.assert_called_once()
        mock_reports_instance.generate.assert_called_once()

    @patch('pyfmto.utilities.cli.ConfigLoader')
    def test_default_config_file(self, mock_loader):
        test_args = ['pyfmto', 'run']
        with patch.object(sys, 'argv', test_args):
            main()
        mock_loader.assert_called_once_with(config='config.yaml')

    def test_show_command(self):
        tmp_dir = Path('temp_dir_for_test')
        algs = ['ALG1', 'ALG2']
        probs = ['PROB1', 'PROB2']
        gen_code('algorithms', algs, tmp_dir)
        gen_code('problems', probs, tmp_dir)
        conf_file = gen_config(
            f"""
            launcher:
                sources: [{tmp_dir}]
                algorithms: [{algs[0]}, {algs[1]}]
                problems: [{probs[0]}, {probs[1]}]
            """,
            tmp_dir
        )
        show_options = {
            'prob': algs,
            'report': list_report_formats(),
            'alg': probs,
            'invalid': ['Invalid']
        }
        for grp, lst in show_options.items():
            for option in lst:
                with self.subTest(grp=grp, option=option):
                    test_args = ['pyfmto', 'show', f'{grp}.{option}', '-c', f'{conf_file}']
                    with patch.object(sys, 'argv', test_args):
                        main()
        with patch.object(sys, 'argv', ['pyfmto', 'show', 'no_dot_in_name']):
            main()

        self.restore_sys_env()
        self.save_sys_env()
        list_options = ['algorithms', 'problems', 'reports', 'invalid']
        args_lst = [['pyfmto', 'list', f'{option}', '-c', f'{conf_file}'] for option in list_options]
        for test_args in args_lst:
            with self.subTest(test_args=test_args):
                with patch.object(sys, 'argv', test_args):
                    main()
        self.delete(tmp_dir)
        self.restore_sys_env()



import os
import platform
import subprocess
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Literal

from rich.console import Console
from rich.table import Table, box
from tabulate import tabulate

from .loggers import logger

__all__ = [
    'add_sources',
    'clear_console',
    'colored',
    'get_cpu_model',
    'get_os_name',
    'get_pkgs_version',
    'matched_str_head',
    'print_dict_as_table',
    'redirect_warnings',
    'tabulate_formats',
    'terminate_popen',
    'titled_tabulate'
]


class TabulatesFormats:
    plain = 'plain'
    simple = 'simple'
    grid = 'grid'
    simple_grid = 'simple_grid'
    rounded_grid = 'rounded_grid'
    heavy_grid = 'heavy_grid'
    mixed_grid = 'mixed_grid'
    double_grid = 'double_grid'
    fancy_grid = 'fancy_grid'
    outline = 'outline'
    simple_outline = 'simple_outline'
    rounded_outline = 'rounded_outline'
    mixed_outline = 'mixed_outline'
    double_outline = 'double_outline'
    fancy_outline = 'fancy_outline'
    pipe = 'pipe'
    presto = 'presto'
    orgtbl = 'orgtbl'
    rst = 'rst'
    mediawiki = 'mediawiki'
    html = 'html'
    latex = 'latex'
    latex_raw = 'latex_raw'
    latex_booktabs = 'latex_booktabs'
    latex_longtable = 'latex_longtable'


tabulate_formats = TabulatesFormats()


def add_sources(paths: list[str]):
    import sys
    for p in paths:
        root = Path(p).resolve()
        if not root.exists():
            logger.warning(f"Path '{root}' does not exist.")
            continue
        if str(root.parent) not in sys.path:
            sys.path.append(str(root.parent))


def terminate_popen(process: subprocess.Popen):
    if process.stdout:
        process.stdout.close()
    if process.stderr:
        process.stderr.close()

    process.terminate()
    try:
        process.wait(timeout=5)
    except subprocess.TimeoutExpired:
        process.kill()
        process.wait()


def print_dict_as_table(data: dict[str, list], true_color="green", false_color="red", title=None):
    if not data:
        return
    lengths = {len(v) for v in data.values()}
    if len(lengths) != 1:
        raise ValueError(f"all values must have the same length: {lengths}")
    table = Table(title=title, box=box.ROUNDED, show_lines=True)
    for key in data.keys():
        table.add_column(str(key), max_width=50, justify="center", overflow='fold', vertical='middle')
    num_rows = lengths.pop()
    keys = list(data.keys())
    for i in range(num_rows):
        row = []
        for key in keys:
            val = data[key][i]
            if isinstance(val, bool):
                color = true_color if val else false_color
                row.append(f"[{color}]{val}[/{color}]")
            else:
                row.append(str(val))
        table.add_row(*row)
    console = Console()
    console.print(table)


def colored(text: str, color: Literal['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'reset']):
    color_map = {
        'black': '\033[30m',
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'reset': '\033[0m'
    }

    if color not in color_map:
        raise ValueError(f"Unsupported color: {color}")

    return f"{color_map[color]}{text}{color_map['reset']}"


def titled_tabulate(title: str, fill_char: str, *args, **kwargs):
    title = ' ' + title if not title.startswith(' ') else title
    title = title + ' ' if not title.endswith(' ') else title
    tab = tabulate(*args, **kwargs)
    tit = title.center(tab.find('\n'), fill_char)
    return f"\n{tit}\n{tab}"


def clear_console():
    if platform.system() == 'Windows':
        os.system('cls')
    else:
        os.system('clear')


def matched_str_head(s: str, str_list: list[str]) -> str:
    for item in str_list:
        if item.startswith(s):
            return item
    return ''


@contextmanager
def redirect_warnings():
    import warnings
    orig_show = warnings.showwarning

    def redirected_show(message, category, *args, **kwargs) -> None:
        if issubclass(category, UserWarning):
            logger.warning(message)
        else:
            orig_show(message, category, *args, **kwargs)

    warnings.showwarning = redirected_show
    try:
        yield
    finally:
        warnings.showwarning = orig_show


def get_os_name():
    os_name = sys.platform
    if os_name == 'win32':
        return 'Windows'
    elif os_name == 'darwin':
        return 'macOS'
    else:
        return 'Linux'


def get_pkgs_version(packages: list[str]):
    res: dict[str, str] = {}

    for name in packages:
        # 1. importlib.metadata (Python 3.8+)
        try:
            from importlib.metadata import version
            res[name] = version(name)
            continue
        except Exception:
            pass

        # 2. module.__version__
        from importlib import import_module
        try:
            module = import_module(name)
            res[name] = module.__version__
            continue
        except Exception:
            pass

        res[name] = 'unknown'
    return res


def get_cpu_model():
    # 1. try platform.processor()
    model = platform.processor().strip()
    if model and model not in ('', 'arm', 'x86_64'):
        return model

    # 2. macOS fallback
    if sys.platform == "darwin":
        try:
            result = subprocess.run(
                ["sysctl", "-n", "machdep.cpu.brand_string"],
                capture_output=True,
                text=True,
                check=True
            )
            model = result.stdout.strip()
            if model:
                return model
        except Exception:
            pass

    # 3. Linux fallback
    if sys.platform.startswith("linux"):
        try:
            with open("/proc/cpuinfo") as f:
                for line in f:
                    if "model name" in line:
                        return line.split(":", 1)[1].strip()
        except Exception:
            pass

    # 4. final fallback
    return "Unknown CPU"
